:- module(curryast, [split_rule/3, check_tree/1, make_name/2, get_params/2,
	             param_at/3, get_function/2]).

% We use high-order prolog.
%:- use_package(hiord).

:- use_module(library(terms), [atom_concat/2] ).

:- include(ast_ops).

% check_tree(AST), takes a Curry AST and verifies that is correct.

check_tree(builtin(_)).

check_tree(Head := Body) :-
	check_head(Head),
	check_expr(Body).

check_head(_Name@Pattern) :-
	%% check_name
	check_pattern(Pattern).

%%% Expressions can be any of:
%%%    - Constants.
%%%    - Function/Cons@Arguments.
%%%    - lambda-Expressions.
 
check_expr(const(_, _)).
check_expr(freevar(_)).
check_expr(param(_)).

check_expr(_FName@ArgsList) :-
	%% check_func(_Func),
	check_expr_list(ArgsList).

check_expr(lambda(Pattern, Expr)) :-
	check_pattern(Pattern),
	check_expr(Expr).

check_expr_list([]).
check_expr_list([E|Es]) :-
	check_expr(E),
	check_expr_list(Es).

check_pattern(P) :-
	list(P).

% split_rule(Rule,Head,Body):-
% Toma una regla y la divide en Cabeza y Cuerpo

split_rule(Rule, Head, Body) :- 
	Rule = Head := Body.

get_params(_@Params, Params).

%% We should handle constants.
get_function(F@_, F) :- !.
get_function(R, R).

param_at(FunApp, [], FunApp).
param_at(_@PList, [Pos|Poss], Exp) :-
	nth(Pos, PList, TExp),
	param_at(TExp, Poss, Exp).

make_name(const(_, C), C).
make_name(qid('Prelude', '[]'), []).
make_name(qid('Prelude', ':'),  .).	
%make_name(qid('Prelude', ','),  ',').

make_name(qid(Module, Name), SName) :-
	atom_concat([Module, '_', Name], SName).

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

