%% deftree.pl: Se pretende que sea un sustituto del actual deftree.pl. 
%% Su labor sera la misma que la del actual con la diferencia de 
%% que tomara su informacion directamente de la salida del parser.
%% (Se elimina la necesidad de la representacion pat para 
%% las reglas y los tipos).

:- module(deftree,	 [defTree/3, call_defTree/2] ).

% Module use.

:- use_module(symtab,      [get_rules/2, get_object_property/3, object_arity/2] ).
:- use_module(scope,       [lookup_symbol/2]).
:- use_module(const_expr,  [is_constant/1] ).
:- use_module(curryast,    [split_rule/3, param_at/3, get_function/2, get_params/2] ).

%

:- use_module(library(sort),       [sort/2] ).
:- use_module(library(lists),      [append/3, length/2, select/3, reverse/2]).
:- use_module(library(aggregates), [findall/3]).

:- include(ast_ops).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Aux functions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Esto es un takeN
comienzos(0, []):-!.
comienzos(N, [[N] | R]):-
	N > 0,
	N1 is N-1,
	comienzos(N1, R).

%% Pertenencia a lista con backtracking.
in(X,[X|_]).
in(X,[_|Xs]) :- 
	in(X,Xs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Pivot calculation.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

mapAtPos([],_,[]).
mapAtPos([R|Rs], Pos, [T|Ts]) :-
	split_rule(R, Head, _),
	param_at(Head, Pos, T),
	mapAtPos(Rs, Pos, Ts).

uniform(Reglas,Pos) :-
	mapAtPos(Reglas, Pos, Terms),
	\+ in(param(_), Terms),
	\+ in(freevar(_), Terms).

% Select the pivot by getting the first position that has not free variables 
% in it. Be warned that this select all the possible pivots on backtracking.
pivote(Reglas, Poss, Piv):-
	in(Piv, Poss),
	uniform(Reglas, Piv).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Split3 calculation.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Aux predicates for traduce.
posiciones(0,[]):-!.
posiciones(N,[N|Ns]):-
	N1 is N-1,
	posiciones(N1,Ns).

prepend(_,[],[]).
prepend(Pos,[Q|Qs],[PQ|PQs]):-
	append(Pos,[Q],PQ),
	prepend(Pos,Qs,PQs).

%%
distrib_rules(Reglas, Con, Pos, R):- % Parser -> Parser
	in(R, Reglas),
	split_rule(R, Head, _),  
	param_at(Head, Pos, Term),
	get_function(Term, Con_aux),
	Con = Con_aux. 

%% 
traduce([], _, _, []).
traduce([Con-Arity|Casos], Reglas, Pos, [(Con, ReglasCon, NPosPos)|Subs]) :-
	findall(
		   R, 
		   distrib_rules(Reglas, Con, Pos, R), 
		   ReglasCon
	       ),
	posiciones(Arity, NPos),
	prepend(Pos,NPos,NPosPos),
	traduce(Casos,Reglas,Pos,Subs).

%% lista_nums/3: Elabora una lista [n1-0, n2-0,...] donde los ni son los
%% numeros contenidos en las reglas que se le pasan y 0 representa la aridad de
%% las constantes numericas
lista_nums([],Ac,Ac) :- 
	!.

lista_const([T|Ts],Ac,Nums) :- 
	!,
	lista_const(T,Ac,N1),
	lista_const(Ts,N1,Nums).

lista_const(Const, Ac, [Const-0|Ac]) :-
	is_constant(Const),
	\+ member(Const-0, Ac),
	!.

lista_const(Rule, Ac, Ac1):-
	split_rule(Rule, Head, _),
	!,
	lista_const(Head, Ac, Ac1).

lista_const(Term, Ac, Nums):-
	get_params(Term, Args),
	!,
	lista_const(Args,Ac,Nums).

lista_const(_,Ac,Ac).

%% Finds all possible cases for a parameter. 
%% This effectively gives us the number of childs of the deftree.

casos(_, Casos, Con) :-
	get_object_property(Con, constructs, Type),
	!,
	findall(Cons-Arity,
	    (
		get_object_property(Cons, constructs, Type),
		object_arity(Cons, Arity)
	    ), 
	    Casos).

%% When we split by a constant, construct the list 
%% of them for the tree.
casos(Reglas, NumsReglas2, X) :-
	is_constant(X),
	!,
	lista_const(Reglas, [],NumsReglas),
	reverse(NumsReglas, NumsReglas2).

% Comment region
casos(_, [const-1], const(_,_)) :-
	!.


%% split3(Reglas,Pos,SubCasos).
%% split3 returns a list of 
split3([Regla| Reglas], Pos, Subs) :-
	split_rule(Regla, Head, _),
	param_at(Head, Pos, PivotTerm),
	get_function(PivotTerm, Pivot),
        casos([Regla|Reglas], Casos, Pivot),
	traduce(Casos, [Regla|Reglas], Pos, Subs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Split3 calculation.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% defTree(Reglas, Posiciones, Arbol)

defTree([],_,fallo-[]).

%% The main case is a branch or rule..
defTree(Reglas, Poss, branch(Piv, Subs)-Ns):-
	pivote(Reglas, Poss, Piv),
	Reglas = [_|_],
	!,
	fake = Ns,  
	split3(Reglas, Piv, Casos), 
	trans(Casos, Piv, Poss, SubAux),
	mapDefTree(SubAux, Subs).

defTree([Regla],Poss,regla-[N]):-
	\+(pivote([Regla],Poss,_)),
	Regla = N.

%% Beware parallel trees maybe broken by now.
defTree(Reglas,Poss,or(DefTrees)-Ns):-
	\+ pivote(Reglas,Poss,_),
	Reglas=[_,_|_],
	Reglas = Ns,
	map_regla_defTree(Reglas,Poss,DefTrees).

map_regla_defTree([],_,[]).
map_regla_defTree([Regla|Reglas],Poss,[DefTree|DefTrees]) :-
	defTree([Regla], Poss, DefTree),
	map_regla_defTree(Reglas, Poss, DefTrees).


mapDefTree([],[]).
mapDefTree([(Con,Reglas,Poss)|Ss],[(Con,T)|Ts]) :-
	defTree(Reglas, Poss, T),
	mapDefTree(Ss, Ts).

trans([],_,_,[]).
trans([(Con,Reglas,NPoss)|Casos],Piv,Poss,[(Con,Reglas,NNPoss)|NCasos]) :-
	append(Antes, [Piv|Despues], Poss),
	append(NPoss, Despues, Aux),
	append(Antes, Aux, NNPoss),
	trans(Casos, Piv, Poss, NCasos).

%%%%%%%%%%%%%%%%%%%%% Arboles Paralelos %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

call_defTree(Pred, Arbol):-
	conjuntos_reglas(Pred,Conjuntos_reglas),
	length(Conjuntos_reglas, L),
	L > 1,
	!,
	object_arity(Pred, Arity),
	comienzos(Arity, Comienzos),
	findall(A,(
		      member(Conjunto,Conjuntos_reglas),
		      Conjunto = Reglas,
		      defTree(Reglas,Comienzos,A)
		  ),
		  Arboles),
	get_rules(Pred, Ns),
	sort(Arboles, ArbolesOrd),
	Arbol=or(ArbolesOrd)-Ns.

call_defTree(Pred,Arbol):-
	object_arity(Pred, Arity),
	!,
	comienzos(Arity, Comienzos),
	get_rules(Pred, Reglas),
	defTree(Reglas,Comienzos,Arbol).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% select(X,[X|Resto],Resto).
% select(X,[Y|Resto],[Y|Z]):-select(X,Resto,Z).

varX_X(var(X),Y,Dicc):-member(X/Y,Dicc),!.
varX_X(Term,TermX,Dicc):-
	Term=..[Id|Args],!,
	varX_X_list(Args,ArgsX,Dicc),
	TermX=..[Id|ArgsX].
varX_X(X,X,_).

varX_X_list([],[],_).
varX_X_list([T|Ts],[Tx|Txs],Dicc):-
	varX_X(T,Tx,Dicc),
	varX_X_list(Ts,Txs,Dicc).

cociente_reglas(Pred,Cociente):-
	% Encuentra todas las Id de regla para pred
        get_rules(Pred, Rules),
	% Las mete a todas las reglas en un lista.
        findall([Id],member(Id,Rules),Cociente0),
	cociente_reglas_aux(Pred,Cociente0,Cociente).

cociente_reglas_aux(_Pred, [], []) :- !.
cociente_reglas_aux(_Pred, [Conj], [Conj]) :- !.
cociente_reglas_aux(Pred, Cociente_ini, Cociente) :-
	% Coge con backtraking una lista de regla
	select(CR1, Cociente_ini, Cociente_ini1),
	select(CR2, Cociente_ini1, Cociente_ini2),
	CR1 = [R1|_],
	CR2 = [R2|_],
	split_rule(R1, H1, _), get_params(H1, P1),
	split_rule(R2, H2, _), get_params(H2, P2),
	varX_X_list(P1, P1v, Dicc),
	varX_X_list(P2, P2v, Dicc),
	P1v = P2v,
	!,
	append(CR1, CR2, Conj1_2),
	append([Conj1_2], Cociente_ini2, Cociente1),
	cociente_reglas_aux(Pred, Cociente1, Cociente).

cociente_reglas_aux(_,Cociente,Cociente).


%%%%%%% Arboles paralelos %%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% uno_de_cada(Conjuntos,Conjunto) toma una lista de listas y devuelve con 
% backtracking conjuntos con un elemento de cada uno de los Conjuntos

uno_de_cada([],[]).
uno_de_cada([Lista|Listas],[Elemento|Elementos]):-
	member(Elemento,Lista),uno_de_cada(Listas,Elementos).

% conjuntos_reglas(Predicado,Conjuntos) toma el nombre de un predicado y devuel-
% ve todos los conjuntos de reglas posibles con los que se puede formar un
% arbol determinista. Las reglas se indican por su numero.

conjuntos_reglas(Pred, Conjs):-
	cociente_reglas(Pred,Cociente),
	findall(C,uno_de_cada(Cociente,C),ConjsAux),
	quitar_repetidos(ConjsAux,[],_,ConjsRev),  
	reverse(ConjsRev, Conjs).


quitar_repetidos_lista([],Ac,Ac,[]).
quitar_repetidos_lista([R|Rs],AcIn,AcOut,[R|Rs2]):-
	\+member(R,AcIn),!,
	quitar_repetidos_lista(Rs,[R|AcIn],AcOut,Rs2).
quitar_repetidos_lista([_R|Rs],AcIn,AcOut,Rs2):-
	quitar_repetidos_lista(Rs,AcIn,AcOut,Rs2).

quitar_repetidos([],Ac,Ac,[]).
quitar_repetidos([Lista|Listas],AcIn,AcOut,[Lista2|Listas2]):-
	quitar_repetidos_lista(Lista,AcIn,AcIn2,Lista2),
	quitar_repetidos(Listas,AcIn2,AcOut,Listas2).


%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

