:- module(lexer_compat, [init_scanner/1, init_scanner_string/1, token_have/1, token_look/1, 
	                 token_back/0, token_skip/0, last_token/1]).

%
:- use_module(library(dynamic), [asserta/1, retract/1, retractall/1]).

:- use_module(lexer, [scanFile/2, scanTokensExpr/2] ).

:- dynamic tokens/1.
:- dynamic last_token/1.

% Compatibility module.

% Now init_scanner takes an argument, the string to 
% lex by now.
% It works lexing the string and storing it as a predicate
% that token_have will return step by step.

init_scanner(File) :-
	retractall(tokens(_)),
	scanFile(File, Tokens),
	asserta(tokens(Tokens)),
	asserta(last_token([(bof, (0,0))])).

init_scanner_string(String) :-
	retractall(tokens(_)),
	scanTokensExpr(String, Tokens),
	asserta(tokens(Tokens)),
	asserta(last_token([(bof, (0,0))])).

%
%  token_have(Token)
%
%  token_have(Token), gets a token from tokens and updates
%  it.

token_have(Token) :-
	retract(tokens([(Token, _Pos)|TR])),
	asserta(tokens(TR)),
	retract(last_token(LTL)),
	asserta(last_token([(Token, _Pos)|LTL])),
	!.

%
%  token_look(Token)
%
%  token_look/1 Token is the next token in the input (used with next).
%

token_look(Token) :-
	tokens([(Token, _Pos)|_]).

%
% skip(Token)
%
% skip/1  skip the next token in the input.
%

token_skip :-
	retract(tokens([T|TR])),
	asserta(tokens(TR)),
	retract(last_token(LTL)),
	asserta(last_token([T|LTL])).

token_back :-
	!,
	retract(tokens(Tks)),
	retract(last_token([LT|LTL])),
	asserta(last_token(LTL)),
	asserta(tokens([LT|Tks])).

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

