:- module(modules, [start_module/1, end_module/1, add_exports/1, load_module/1, 
	            current_mod_name/1] ).

%%
:- use_module(library(lists),      [append/3] ).
:- use_module(library(dynamic),    [assert/1, retract/1] ).
:- use_module(library(aggregates), [findall/3] ).

%% 
:- use_module(symtab,            [get_rules/2, get_object_property/3, get_constructors/2,
	                          set_object_property/3, load_interfaces/1] ).

%% This module implements some predicates for the module system.

:- dynamic current_module/2.

%% A module is represented like:
%%
%% curry_mod(Name, Symbols)

%% 

current_mod_name(Name) :-
	current_module(Name, _).

%% Predicates for parser.

start_module(Module) :-
	assert(current_module(Module, [])).

%% Resolve true exports and record them in the symbol table.
end_module(Module) :-
	retract(current_module(Module, Exports)),
	(
	    Exports = all ->
	    findall(id(F), (get_rules(F, Rules), \+ Rules = [builtin(imported)]), NExports),
	    findall(type(qid(Module, Name), Cons), get_constructors(qid(Module, Name), Cons) , TExports),
	    append(TExports, NExports, FExports),
	    set_object_property(Module, exports, FExports)
	;
	    set_object_property(Module, exports, Exports)
	).

%% Warning, the condition inside the findall should be handled with care.
load_exports(Module, Imports) :-
	use_module(Module),
	findall(iface(Iface), 
	        ( 
		    Module:exports(Iface),
		    (
			Iface = id(qid(_, Name), _, _),
			(
			    member(id(Name), Imports),
			    !			    
			;
			    Imports = []
			)
		    ;
			Iface = type(Name, _)
		    )
		), Exports),
	load_interfaces(Exports).

%% This is called when a import is found.
load_module(module(Name, Imports, _Modifiers)) :-
	load_exports(Name, Imports),
	set_object_property(dummy, imports, module(Name, Imports, _Modifiers)),
	!.
	

%%
add_exports(all) :-
	retract(current_module(M, _)),
	assert(current_module(M, all)).

add_exports(List) :-
	list(List),
	retract(current_module(M, L)),
	append(List, L, NewL),
	assert(current_module(M, NewL)).
	
%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

