:- module(typer, [check_infer_types/0]).

% Librerias de Sloth

% :-include(operators).
%  :-ensure_loaded(dynamics).


 % :-use_module(sicstus).
 :-use_module(symtab).
 % :-use_module(trad).
 % :-use_module(aux).
 % :-use_module(scanner).
 % :-use_module(typer).
 % :-use_module(deftree).
 :-use_module(parser). % OJO!! el lambda-lifting no va

:- include(type_ops).
:- include(ast_ops).

% librerias que se usan para este módulo

:-use_module(library(pretty_print)).
:-use_module(library(lists)).
:-use_module(library(terms_vars)).


% definimos los predicados dinámicos tipot y reglat
:- dynamic(tipot/2).
:- dynamic(reglat/3).

% egallego@babel: 15 feb 2005, 18:34
% Dummy typer checker.
check_infer_types.


%       Extraemos los constructores, data y reglas de la tabla
%       ------------------------------------------------------


% listatipos saca los tipos de los constructores en una lista
%listatipos(Rs):- findall((X,Y),lookup(X,type,Y),Rs).
listatipos(Rs):- findall((X,Y),get_object_property(X,type,Y),Rs).


% instipo guarda los tipos de los constructores a partir de listatipos
instipo([]).
instipo([(A,B)|Xs]):- assert(tipot(A,B)), instipo(Xs).


% Tablareglas obtiene la lista de reglas de la tabla de Simbolos (TS)

%tablareglas(L):- findall(X,lookup(_,abstree,X),L).
tablareglas(R):- findall((A,X),get_rules(A,X),L), listaReglas(L,R).



% convierte la lista de reglas obtenida por tablareglas(L) en una
% unica lista de arboles sintacticos.
listaReglas([],[]).
listaReglas([X|Xs],Rs):- lreglas2(X,A1), listaReglas(Xs,A2), append(A1,A2,Rs).

lreglas2((_,[builtin]),[]):- !.
lreglas2(('@Objective',_),[]):- !.
lreglas2((A,B),B).


% insregla transforma la lista de reglas obtenidas con tablareglas
% al formato adecuado para ser tipadas, y las guarda.

insregla([]).

insregla([X|Xs]):- X=(A:=B),  A=(A1@R1), verite(B,B1),
                   l2var(regla(A1,R1,B1),Reg,Lvar),
                   annadevars(Lvar,Lvar2,W),
                   sust9(Lvar2,Reg,Reg2),
                   Reg2=regla(RR1,RR2,RR3),
                   assert(reglat(RR1,RR2,RR3)), insregla(Xs).
% por si la regla no respeta el formato.
insregla([X|Xs]):- X\=(A:=B), insregla(Xs). 

% verite comprueba si se trata de un if-then-else (ite).
verite((X->Y),ite(X,A,B)):- Y=(A#B).
verite(X,X):- X\=(A->B).


ini(X):- listatipos(Rs), instipo(Rs), % obtiene los tipos predefinidos
             tablareglas(L), insregla(L), % obtiene las reglas de la TS
             todaslasreglas(X). % Calcula los tipos de las reglas


% el predicado iniciar(X) lo hace todo.
iniciar(X):- listatipos(Rs), instipo(Rs), % obtiene los tipos predefinidos
             tablareglas(L), insregla(L), % obtiene las reglas de la TS
             todaslasreglas(X), % Calcula los tipos de las reglas
             gtipotabla(X). % Guarda las reglas en la TS
             

% para sacar la lista de tipos de las reglas, sin guardarlas en la tabla
l_tipos(X):-listatipos(Rs), instipo(Rs), % obtiene los tipos predefinidos
             tablareglas(L), insregla(L), % obtiene las reglas de la TS
             todaslasreglas(X), % Calcula los tipos de las reglas
             verErrores(Errores). % Muestra la lista de errores


/*
l_tipos2(X):-   % para CLASES
             update(==,type,(eq(_G)=>(_G->_G->'Bool'))),
             listatipos(Rs), instipo(Rs), % obtiene los tipos predefinidos
             tablareglas(L), insregla(L), % obtiene las reglas de la TS
             todaslasreglas(X), % Calcula los tipos de las reglas CON CLASES
             verErrores(Errores).
  
*/       

%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%  PREDICADO TIPO
%%%%%%%%%%%%%%%%%%%%%%%%%%




%                     Tipos básicos predefinidos
%                     -------------------------

% tipo Float
  tipo(X,'Float') :- float(X), !.

% tipo Int
  tipo(X,'Int') :- int(X), !.

% tipo List
  %  tipo([],[_]).  ya esta en la TS

    tipo([X|[]],list(Y)) :- tipo(X,Y) ; error_regla(2). 
    tipo([X|Xs],list(Y)) :- tiplista(X,Y), Xs\=[], tipo(Xs,list(Y)).

    tiplista(X,Y):- tipo(X,Y) ; error_regla(2).    

% tipo tupla


    tipo((X,Y),(A,B)) :- tipo(X,A), tipo(Y,B).


 %                 tipos derivados de la tabla de simbolos   
 %                 ---------------------------------------

 tipo(num(N),X):- integer(N),X='Int',! ; X='Float',!.
 tipo(char(_),'Char').
 tipo(str(_),['Char']).




% CLASES!!!!

% Comparar_clases

clases_predefinidas(X):- X=[(eq,Eq),(show,Show), (read,[]), (bounded,[]),
                           (monad,Monad),(functor,[]), (enum,Enum)],
                  Eq=[(ord,Ord),(num,Num)],
                  Show=[(num,Num)],
                  Ord=[(real,Real)],
                  Num=[(real,Real),(fractional,Fractional)],
                  Enum=[(integral,[])],
                  Real=[(integral,[]),(realFrac,RealFrac)],
                  Fractional=[(realFrac,RealFrac),(floating,Floating)],
                  RealFrac=[(realFloat,[])],
                  Floating=[(realFloat,[])],
                  Monad=[(monadPlus,[])].


comparar_clases(X,X,X):- !.                                      
comparar_clases(X,Y,Y):- subclases(X,L), esta_clase2(Y,L), !. 
comparar_clases(X,Y,X):- subclases(Y,L), esta_clase2(X,L), !.
comparar_clases(X,Y,Z):- comparten_clase(X,Y,Z), !.

esta_clase(A,[]):- fail.
esta_clase(A,[(A,Lb)|S]).
esta_clase(A,[(B,Lb)|S]):- A\=B, esta_clase(A,S). 

esta_clase2(A,[]):- fail,!.
esta_clase2(A,[(A,Lb)|S]):- !.
esta_clase2(A,[(B,Lb)|S]):- A\=B, (esta_clase2(A,S),!) ;
                                  (esta_clase2(A,Lb),!).

comparten_clase(X,Y,Z):- subclases(X,Lx), subclases(Y,Ly), comparten2(Lx,Ly,Z).

comparten2([(Z,La)|S],Ly,Z):- esta_clase2(Z,Ly), !.
comparten2([(A,La)|S],Ly,Z):- (comparten2(S,Ly,Z),!) ;
                              ( comparten2(La,Ly,Z),!) . 


subclases(A,L):- clases_predefinidas(C), esta_clase2(A,C),
                                      lista_subclases(A,L,C).

lista_subclases(A,L,[(A,L)|S]):- !.
lista_subclases(A,L,[(B,Lb)|S]):- A\=B, esta_clase2(A,Lb),
                                       lista_subclases(A,L,Lb), !. 
lista_subclases(A,L,[(B,Lb)|S]):- A\=B, esta_clase2(A,S),
                                       lista_subclases(A,L,S), !.




                  
                  



% Multifiles verify_attribute/2 y combine_attribute/2

:- multifile combine_attributes/2.
combine_attributes(V1,V2) :- V1=V2. %aqui debe ir comparar_clases

:- multifile verify_attribute/2.
verify_attribute(A,T) :-  A=(X,Y), X=eq(V), detach_attribute(V), V=T.
  % hay que sacar la variable V de otra manera, ya que puede haber
  % varias clases ademas de eq.


% deben ponerse aqui todas las clases que se usan?
eq(X):- true.
ee(X,Y):- varset(X,Y).





%                             Reglas de inferencia
%                             --------------------

% los tipos sacados de la tabla de simbolos
  tipo(A,T):- tipot(A,T), !.

% Variables
  tipo(var(X),X).

% condicional
    tipo(ite(A,B,C),T) :- ite1(A), ite2(B,C,T).

    ite1(A):-  tipo(A,'Bool') ; error_regla(7).

    ite2(B,C,T) :- tipo(B,T), tipo(C,T)
			       ; error_regla(6).

% abstracciones
    tipo(lambda(P,E),(T1->T2)):- mis1(P,T1), mis1(E,T2).

    % para comprobar errores type Mismatch
    mis1(P,T1):- tipo(P,T1) ; error_regla(3).



    tipo(case([]),X).
    tipo(case([L1|Ls]),(T1->T2)):-  mis1(L1,(T1->T2)), tipo(case(Ls),(T1->T2)).

    

% Aplicaciones
 %   tipo((F @ X),T2) :- tipo(F,(T1->T2)), mis1(X,T1).  % MODIF para MiSMATCH

 tipo((F @ X),T2) :- (tipo(F,(T1->T2)), mis1(X,T1),! )  % MODIF para MiSMATCH
                 ;( tipo(F,(=>(eq(A),(T1->T2)))),
                   attach_attribute(A,(eq(A),Att)),
                    tipo(X,T1) ).

% Let
  tipo(le(X,E1,E2),T) :- tipo(E1,T1), tipo(X,T1), tipo(E2,T).

% Fixed Point
  tipo(fix(X,E),T) :- tipo(X,T), tipo(E,T).




% tipador de reglas

%        convierte las reglas en una sola expresión con lambdas.
%        ------------------------------------------------------


 
 elambdaux(N,S,[],S).
 elambdaux(N,lambda(X,E),[X|Xs],S):- elambdaux(N,E,Xs,S).

 elambda(N,E):-regla(N,L,S), elambdaux(N,E,L,S).
 

% modificamos ecase para que sólo ponga case si hay varias reglas

 ecase(N,case(L)):- setof(X,elambda(N,X),L), \+ hayuna(L).
 ecase(N,X):- setof(X,elambda(N,X),L), hayuna(L),L=[X|[]].

lfix([],[]).
lfix([X|Xs],[Y|R]):- ecase(X,Y), lfix(Xs,R).


% saca la expresión con fix de una componente conexa
% y  sustituye los nombre de reglas por var(X) en los fix,
%  y lo convierte en tuplas.
ffix2(L,L1):- hayuna(L), L=[A], \+esrecur(A), ecase(A,L1),!.
ffix2(L,R):- hayuna(L),L=[A], esrecur(A), ecase(A,B), sust(A,fix(A,B),R),!.
ffix2(L,fix(X,Y)):- \+hayuna(L), lfix(L,L1),
   reemplazo(L,fix(L,L1),R),R=fix(A,B), listatupla(A,X), listatupla(B,Y),!.


/*
% obtiene el tipo de una lista de reglas
tipear([],[]).
tipear([X|Xs],[W,L1]):- ffix2(X,E), X=[Actual], 
                              update('@Type_data',regla_actual,Actual),
			       tipoerror2(E,T), % 2 clases 1 sin clases
                               gua(X,T,W), tipear(Xs,L1).

*/
tipear([],[]).
tipear([X|Xs],[W,L1]):- ffix2(X,E), tipoerror(E,T), gua(X,T,W), tipear(Xs,L1).



% si la expresión es no tipable, se le pone el tipo No Tipable.
tipoerror(E,T):- tipo(E,T),! ; T='No Tipable',!.

% PARA CLASES

tipoerror2(E,T2):- tipo(E,T), pclas(T,T2), ! ;    T2='No Tipable',!.


% obtiene la lista de variables de un tipo T y añade las clases, si las
% variables tenian atributos asociados.
pclas(T,T2):- varset(T,L), ponerclases(T,L,T2).


% a partir de un tipo T y de una lista de variables [A|As]
% añade las clases Eq(X)=> X etc.. al tipo original en T2
ponerclases(T,[],T).
ponerclases(T,[A|As],T2):- (get_attribute(A,(A1,A2)),
			    ponerclases((A1=>T),As,T2),!) ;
                            (ponerclases(T,As,T2),!).

% HASTA AQUI

% guarda el tipo obtenido por si es necesario para tipear otra regla
gua([A|B],(X,Y),[(A,X),Rs]):- B\=[], assert(tipot(A,X)), gua(B,Y,Rs).
gua([A|[]],T,(A,T)):- assert(tipot(A,T)),!.


% obtiene el tipo de todas las reglas y lo guarda en la lista T
todaslasreglas(T):- ordeneval(L), tipear(L,L1),li(L1,T).


% para guardar la lista de tipos de reglas en la tabla de simbolos      

gtipotabla([]).
gtipotabla([X|Xs]):- X=(A,B),
                  % hay que comprobar que no esta ya insertado????
                     insert(A,type,B), gtipotabla(Xs).

% obtiene el tipo de una regla (se usa para ejemplos)
% tiporegla(X,T):- cconexa(X,L), ffix2(L,E), tipo(E,T), pretty_print(T,[]).
 tiporegla(X,T):- cconexa(X,L), ffix2(L,E), tipo(E,T).



%  predicados para convertir var(x) en var(X) en las reglas         
% ---------------------------------------------------------

  
  % unir une dos listas sin elementos repetidos

   unir([],[],[]):- !.
   unir([],Y,Y):- !.
   unir(X,[],X):- !.
   unir([X|Xs],Y,[X|W]):- noesta(X,Y), unir(Xs,Y,W).
   unir([X|Xs],Y,W):- \+ noesta(X,Y), unir(Xs,Y,W).
   
  % para usar despues de sacar la lista de reglas
   annadevars([],[],var(O)).
   annadevars([var(N)|Rs],[W|Zs],(Y,L)):- copy_term(var(X),Y),
                                      W=(var(N),Y),                                                                 annadevars(Rs,Zs,L). 


  % l2var coge una regla y la transforma en regla valida
  % y saca la lista de variables que tiene.

   l2var(regla(A,B,C),regla(X,Y,Z),L):- l2var(A,X,L1),
                    l2var(B,Y,L2), l2var(C,Z,L3), unir(L1,L2,L4),
                    unir(L4,L3,L).
   l2var(id(A),A,[]).
   l2var(var(A),var(A),[var(A)]).


   l2var((A@B),(X@Y),L):- \+list(B) , 
                       l2var(A,X,L1), l2var(B,Y,L2), unir(L1,L2,L).


%% ERROR. El problema es que, al llamar a l2var despues de l2varaux
%% vuelve a ejecutarse l2varaux y el tipo a@[b] no aparece nunca.
%% aparece a@b

   l2var((A@B),X,L):- list(B),l2varaux((A@B),C), l2var2(C,X,L).

  
 %% modificado

   

   

  % para convertir las expresiones A@[ B1 B2 ] en expresiones tipables
  % como ((A@B1)@B2)

   

   l2varaux((A@[]),A).
  

   l2varaux((A@[B|Bs]), C):- B\=(M@N) ,X=(A@B),
                                  l2varaux((X@Bs),C).

   l2varaux((A@[B|Bs]), C):- B=(M@N), l2varaux(B, B1), X=(A@B1), 
          

               l2varaux((X@Bs), C).
%% NO vale

   l2varaux2((A@[]),A,L).
  

   l2varaux2((A@[B|Bs]), C,L):- B\=(M@N) ,X=(A@B),
                                  l2var((X@Bs),C,L).

   l2varaux2((A@[B|Bs]), C,L):- B=(M@N), l2var(B, B1,L1), X=(A@B1), 
                         l2varaux2((X@Bs), C,L2), unir(L1,L2,L).

    %% Esto no vale                                      
   l2varaux1((A@[]),A).
  

   l2varaux1((A@[B|Bs]), ((A@B)@C)):- B\=(M@N) ,
                                  l2varaux1(Bs,C).

   l2varaux1((A@[B|Bs]), ((A@B1)@C)):- B=(M@N), l2varaux1(B, B1), 
                         l2varaux1(Bs, C).
                                   
  


%% fin modificaciones

   l2var(ite(A,B,C),ite(X,Y,Z),L):- l2var(A,X,L1),
                    l2var(B,Y,L2), l2var(C,Z,L3), unir(L1,L2,L4),
                    unir(L4,L3,L).

/*
% formato de constructores (div y >>= dan error) ¡Arreglarlo!   

   l2var(A + B,((@(+,A))@B),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A - B,((@(-,A))@B),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A * B,((@(*,A))@B),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A / B,((@(/,A))@B),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A==B,((@(==,A))@B),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).

   l2var(A mod B,@(mod,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
%  l2var(A div B,@(div,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A>>B,@(>>,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
%  l2var(A>>=B,@(>>=,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A=B,@(=,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
  
   l2var(A<>B,@(<>,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A<B,@(<,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A>B,@(>,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A=<B,@(=<,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A>=B,@(>=,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A&&B,@(&&,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A'||'B,@('!!',@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
  
*/
 
   l2var((A,B),(X,Y),L):- l2var(A,X,L1),l2var(B,Y,L2),unir(L1,L2,L).
   l2var(A,X,L):- list(A), ll2var(A,X,L),!.
   l2var(num(X),num(X),[]).
   l2var(char(X),char(X),[]).
   l2var(str(X),str(X),[]).
 
   % para usar l2var con listas
   ll2var([],[],[]):- !.
   ll2var([A|As],[X|Xs],L):- l2var(A,X,L1), ll2var(As,Xs,L2), unir(L1,L2,L).



%%%%%%%%%% AUXILIAR DE L2VAR %%%%%%%%%%%% L2VAR2
%%%%%%
%%%  Se usa para dar una segunda pasada después de examinar un formato A@B
%%% sin que vuelva a examinar A@B.
%%% ¡ Se puede mejorar mucho ! (quizas usando una variable auxiliar que
%% indique cuándo se ha dado una segunda pasada en l2varaux).


 % l2var coge una regla y la transforma en regla valida
  % y saca la lista de variables que tiene.

   l2var2(regla(A,B,C),regla(X,Y,Z),L):- l2var2(A,X,L1),
                    l2var2(B,Y,L2), l2var2(C,Z,L3), unir(L1,L2,L4),
                    unir(L4,L3,L).
   l2var2(id(A),A,[]).
   l2var2(var(A),var(A),[var(A)]).


   l2var2((A@B),(X@Y),L):- 
                       l2var2(A,X,L1), l2var2(B,Y,L2), unir(L1,L2,L).


 l2var2(ite(A,B,C),ite(X,Y,Z),L):- l2var2(A,X,L1),
                    l2var2(B,Y,L2), l2var2(C,Z,L3), unir(L1,L2,L4),
                    unir(L4,L3,L).
   

/*

   l2var2(A + B,((@(+,A))@B),L):- l2var2(A,X,L1),l2var2(B,Y,L2), unir(L1,L2,L).
   l2var2(A - B,((@(-,A))@B),L):- l2var2(A,X,L1),l2var2(B,Y,L2), unir(L1,L2,L).
   l2var2(A * B,((@(*,A))@B),L):- l2var2(A,X,L1),l2var2(B,Y,L2), unir(L1,L2,L).
   l2var2(A / B,((@(/,A))@B),L):- l2var2(A,X,L1),l2var2(B,Y,L2), unir(L1,L2,L).
   l2var2(A==B,((@(==,A))@B),L):- l2var2(A,X,L1),l2var2(B,Y,L2), unir(L1,L2,L).

   l2var2(A mod B,@(mod,@(X,Y)),L):- l2var2(A,X,L1),l2var2(B,Y,L2), unir(L1,L2,L).
%  l2var2(A div B,@(div,@(X,Y)),L):- l2var2(A,X,L1),l2var2(B,Y,L2), unir(L1,L2,L).
   l2var2(A>>B,@(>>,@(X,Y)),L):- l2var2(A,X,L1),l2var2(B,Y,L2), unir(L1,L2,L).
%  l2var2(A>>=B,@(>>=,@(X,Y)),L):- l2var2(A,X,L1),l2var2(B,Y,L2), unir(L1,L2,L).
   l2var2(A=B,@(=,@(X,Y)),L):- l2var2(A,X,L1),l2var2(B,Y,L2), unir(L1,L2,L).
  
   l2var2(A<>B,@(<>,@(X,Y)),L):- l2var2(A,X,L1),l2var2(B,Y,L2), unir(L1,L2,L).
   l2var2(A<B,@(<,@(X,Y)),L):- l2var2(A,X,L1),l2var2(B,Y,L2), unir(L1,L2,L).
   l2var2(A>B,@(>,@(X,Y)),L):- l2var2(A,X,L1),l2var2(B,Y,L2), unir(L1,L2,L).
   l2var2(A=<B,@(=<,@(X,Y)),L):- l2var2(A,X,L1),l2var2(B,Y,L2), unir(L1,L2,L).
   l2var2(A>=B,@(>=,@(X,Y)),L):- l2var2(A,X,L1),l2var2(B,Y,L2), unir(L1,L2,L).
   l2var2(A&&B,@(&&,@(X,Y)),L):- l2var2(A,X,L1),l2var2(B,Y,L2), unir(L1,L2,L).
   l2var2(A'||'B,@('!!',@(X,Y)),L):- l2var2(A,X,L1),l2var2(B,Y,L2), unir(L1,L2,L).
  
*/

 
   l2var2((A,B),(X,Y),L):- l2var2(A,X,L1),l2var2(B,Y,L2),unir(L1,L2,L).
   l2var2(A,X,L):- list(A), ll2var2(A,X,L),!.
   l2var2(num(X),num(X),[]).
   l2var2(char(X),char(X),[]).
   l2var2(str(X),str(X),[]).
   
  % para usar l2var con listas
   ll2var2([],[],[]):- !.
   ll2var2([A|As],[X|Xs],L):- l2var2(A,X,L1), ll2var2(As,Xs,L2), unir(L1,L2,L).


%%%%% FIN AUXILIAR
%%%%%%%%%%%%%%%%%%%


    % sustituye una lista de variables var(x) por sus var(X) correpondientes
    % en la expresión E
    sust9([],E,E).
    sust9([(A,B)|Xs],E,E2):- sust5(A,E,B,E1), sust9(Xs,E1,E2).

  
   %  Sustituye N por S, se usa para sustituir var(x) por var(X) 
   % en las reglas obtenidas por la TS, para que sean tipables.
    sust5(N,N1,S,S):- N==N1.
    sust5(N,regla(A,B,C),S,regla(X,Y,Z)):- sust5(N,A,S,X), sust5(N,B,S,Y),
                                           sust5(N,C,S,Z).
    sust5(N,ite(A,B,C),S,ite(J,K,L)):- sust5(N,A,S,J),sust5(N,B,S,K),
                                       sust5(N,C,S,L).
    sust5(N,(A@B),S,(J@K)):- sust5(N,A,S,J),sust5(N,B,S,K).
    sust5(N,(A,B),S,(J,K)):- sust5(N,A,S,J),sust5(N,B,S,K).
    sust5(N,B,S,B):- atom(B), \+ list(B), N\=B.
    sust5(N,num(X),S,num(X)).
    sust5(N,char(X),S,char(X)).
    sust5(N,str(X),S,str(X)).
    sust5(N,B,S,B):- number(B).
    sust5(N,var(A),_,var(A)):- \+(N==var(A)).
    sust5(N,L,S,L1):- list(L), sustl5(N,L,S,L1).

    % lo mismo que sust5, por si la expresión E es una lista
    sustl5(N,[],S,[]).
    sustl5(N,[X|Xs],S,[A|As]):- sust5(N,X,S,A), sustl5(N,Xs,S,As).
   


%               Calculo de componentes conexas para fix
%               ---------------------------------------


% predicado para ver si un nombre aparece en una expresion

    esta(N,N).
    esta(N,ite(X,Y,Z)):- esta(N,X),! ; esta(N,Y),! ; esta(N,Z).
    esta(N,lambda(X,Y)):- esta(N,X), ! ; esta(N,Y), !.
    esta(N, (X@Y)):- esta(N,X),! ; esta(N,Y),!.
    esta(N,le(X,Y,Z)):- esta(N,X),!; esta(N,Y),! ; esta(N,Z),!.
    esta(N,(X,Y)):- esta(N,X),! ; esta(N,Y), !.

% predicado para sustituir el nombre de la regla por una var(X)
% y usarlo en el tipado de fix (punto fijo)

    sust(N,N,var(X)).
    sust(N,lambda(A,B),lambda(J,K)):- sust(N,A,J), sust(N,B,K).
    sust(N,ite(A,B,C),ite(J,K,L)):- sust(N,A,J),sust(N,B,K),sust(N,C,L).
    sust(N,(A@B),(J@K)):- sust(N,A,J),sust(N,B,K).
    sust(N,le(A,B,C),le(J,K,L)):- sust(N,A,J),sust(N,B,K),sust(N,C,L).
    sust(N,(A,B),(J,K)):- sust(N,A,J),sust(N,B,K).
    sust(N,B,B):- atom(B), \+ list(B), N\=B.
    sust(N,B,B):- number(B).
    sust(N,var(A),var(A)).
    sust(N,num(X),num(X)).
    sust(N,id(X),id(X)).
    sust(N,char(X),char(X)).
    sust(N,str(X),str(X)).

    sust(N,L,L1):- list(L), sustl(N,L,L1).
    sust(N,case(X),case(Y)):- sust(N,X,Y).
    sust(N,fix(X,Y),fix(A,B)):- sust(N,X,A),sust(N,Y,B).

   % como sust, pero para listas
    sustl(N,[],[]).
    sustl(N,[X|Xs],[A|As]):- sust(N,X,A), sustl(N,Xs,As).


 
    % reemplaza todas las variables en un fix([],[]) (A debe ser fix)    
    reemplazo([],A,A).
    reemplazo([X|Xs],A,Z):- sust(X,A,B), reemplazo(Xs,B,Z).

    reemplazo2(fix(A,B),C):- reemplazo(A,fix(A,B),C).


   
% predicado para obtener la expresión final tipable

   efinal(N,fix(N,case(L))):- elambda(N,Y), esta(N,Y),
                           setof(X,elambda(N,X),L),!.

   efinal(N,case(L)):- setof(X,elambda(N,X),L).

% para ver si una función se llama a si misma
esrecur(X):- regla(X,_,Y), esta(X,Y).

% para calcular funcions que llaman a otras funciones
% apunta(N,E) <-- N apunta a E.

apunta(N,E):- regla(N,_,Y), esregla(E), N\=E,  esta(E,Y).

% apunta, si el segundo término es una lista.
% para ver si una componente conexa apunta a una función).
apunta2(N,[A|[]]):- apunta(N,A).
apunta2(N,[X|Xs]):- apunta(N,X), Xs\=[],! ;  apunta2(N,Xs).

% apunta, si el primer término es una lista.
% (para ver si una funcion apunta a una componente conexa).
apunta3([X|[]],Y):- apunta(X,Y).
apunta3([X|Xs],Y):- apunta(X,Y), Xs\=[], ! ; apunta3(Xs,Y).

% apunta si los dos términos son listas (para ver si dos componentes
% conexas se apuntan)
apunta1([X|[]],[Y|Ys]):- apunta2(X,[Y|Ys]).
apunta1([X|Xs],[Y|Ys]):- apunta2(X,[Y|Ys]), Xs\=[],! ; apunta1(Xs,[Y|Ys]).



%                    Predicados para Listas
% algunas se podrían sustituir por las que están en la libreria list
% (programadas en un momento en el que no se disponia de la documentación
% de la libreria list)
%  -----------------------------------------------


% predicado para convertir una lista de listas en una lista
li([X|Xs],[X|Ys]):- \+ list(X), li(Xs,Ys).
li([X|Xs],R):- list(X), li(X,Y), li(Xs,Ys), append(Y,Ys,R).
li([],[]).
 
 
  % convierte una lista en una tupla
  listatupla([A|As],(A,Bs)):- listatupla(As,Bs).
  listatupla([X|[]],X).

 % indica si un elemento NO ESTA en una lista. (es como nocontainsx(L,X) ).
 noesta(A,[]).
 noesta(A,[X|Xs]):- A\=X, noesta(A,Xs).

 % elimina un elemento de una lista (es como delete(L1,E,L2) )
  quitar(A,[],[]).
  quitar(A,[A|Xs],Ys):- quitar(A,Xs,Ys).
  quitar(A,[X|Xs],[X|Ys]):- A\=X, quitar(A,Xs,Ys).

 % elimina los elementos de una lista de otra lista  (es como difference(Linicial,Lquitar,Lfinal)
  quitarl([],Xs,Xs).
  quitarl([A|As],Xs,Zs):- quitar(A,Xs,Ys), quitarl(As,Ys,Zs).

 % ordena una lista de funciones que se apuntan unas a otras
  ordenar([],Xs,Xs).
  ordenar([A|As],Bs,Zs):- inserta(A,Bs,Cs), ordenar(As,Cs,Zs).

 % inserta una funcion dependiendo de si apunta o no a las demás
  inserta(A,[],[A]).
  inserta(A,[B|Bs],[A|Cs]):- apunta1(B,A), Cs=[B|Bs].
  inserta(A,[B|Bs],[B|Cs]):- \+ apunta1(B,A), inserta(A,Bs,Cs).

% devuelve una lista sin elementos repetidos 
  norep([],[]).
  norep([X|[]],[X|[]]).
  norep([X|Xs],[X|Zs]):- Xs\=[], quitar(X,Xs,Ys), norep(Ys,Zs).

 % devuelve la lista de reglas
  lreglas(R):- findall(X,regla(X,_,_),L), norep(L,R).

% indica si la lista sólo tiene un elemento
  hayuna([X|[]]).



%                Predicados para Reglas
%                ----------------------

  esregla(X):- lreglas(L), pertenece(X,L).
 
  pertenece(X,[X|[]]).
  pertenece(A,[X|Xs]):- A=X ; pertenece(A,Xs).

   

 %funcion para calcular el camino con una lista de grafos

 cam(N,Y):- camaux(N,Y,[]).
 
 camaux(X,X,L).
 camaux(X,Y,L):- apunta(X,Z), noesta(Z,L), camaux(Z,Y,[Z|L]).


 % Cálculo de la componente conexa usando cam
 cconexa2(X,Y):- cam(X,Y), cam(Y,X).

 cconexa(X,L):- setof(Y,cconexa2(X,Y),L).

% calculo de la lista de componentes conexas a partir de la lista de reglas

 lconexa(L):- lreglas(R), lconexa2(R,L).
 lconexa2([],[]).
 lconexa2([X|Xs],[Y|Ys]):- cconexa(X,Y), quitarl(Y,[X|Xs],Zs),
                           lconexa2(Zs,Ys).

% devuelve la lista de componentes conexas ordenadas para su evaluacion

  ordeneval(L):- lconexa(R), ordenar(R,[],L).  






%                             Reglas de la TS
%                            ------------------


   regla(A,B,C):- reglat(A,B,C).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%  TRATAMIENTO DE ERRORES

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%% función para comprobar si el tipo inferido
%% concuerda con el especificado.

%% comprobar_tipos

% si el tipo inferido y el definido son iguales, todo queda igual
% si el tipo inferido es más restrictivo, da error
% si el tipo definido es el más restrictivo, se da un warning 
%  y se usa este tipo.


/*

    comprobar_tipo(Regla,InfrType,DefType) :- 
                   numbervars(DefType,0,_),
                   numbervars(InfrType,0,_),
                   InfrType==DefType,
                   insert(Regla,type,InfrType), !
    ;
                comprobar_tipo2(Regla,InfrType,DefType).
    

 comprobar_tipo2(Regla,InfrType,DefType) :-
                   InfrType=DefType,
                   comprobar_tipo3(Regla,DefType), !
            
    ;
                   typerror(Regla,11).

 comprobar_tipo3(Regla,X) :-
                   lookup(Regla,defType,DefType),
                   numbervars(DefType,0,_),
                   numbervars(X,0,_),
                   DefType==X,
                   type_Warning(Regla,13),
                   insert(Regla,type,X)
                    
    ;
                   typerror(Regla,11).
                   

% Comprueba la lista de tipos inferidos con los tipos definidos
    comprobar_tipos([]).
    comprobar_tipos([X|Xs]):- X=(Regla,InfrType),
                              InfrType\='No Tipable',
                              lookup(Regla,defType,DefType), 
                              comprobar_tipo(Regla,InfrType,DefType),
                              comprobar_tipos(Xs), !
                           ; comprobar_tipos(Xs).



    mensaje_error(_,Num_Error,Mensaje):-

% Lista de errores de tipos
               Num_Error=1, Mensaje='Undefined Symbol'
    ;          Num_Error=2, Mensaje='Different types in list'
    ;          Num_Error=3, Mensaje='Type Mismatch'
    ;          Num_Error=4, Mensaje='Undefined Constructor'
    ;          Num_Error=5, Mensaje='A constraint type was expected'
    ;          Num_Error=6, Mensaje='Then and Else expressions have
different types'
    ;          Num_Error=7, Mensaje='If expression must be boolean'
    ;          Num_Error=8, Mensaje='Invalid constraint in choice expression'
    ;          Num_Error=9, Mensaje='Expression in choice are not the same type'
    ;          Num_Error=10, Mensaje='Rules have different types'
    ;          Num_Error=11, Mensaje='Inferred type does not match the specified one'
    ;          Num_Error=12, Mensaje='Infinite type for function'

% A partir de aqui van los warnings

    ;          Num_Error=13, Mensaje='Inferred type does not match the specified one'
    ;          Num_Error>12, Mensaje=0
    ;          Num_Error<1,  Mensaje=0.


error_regla(Numero):-
    lookup('@Type_data',regla_actual,Regla), typerror(Regla,Numero), fail.


%%%%% Guardan los errores y Warnings en la tabla de simbolos

    typerror(Regla,Num_Error):-
               mensaje_error(Regla,Num_Error,Mensaje),
               insert('@Type_Error',Regla,Mensaje).

    type_Warning(Regla,Num_Error):-
               mensaje_error(Regla,Num_Error,Mensaje),
               insert('@Warning',Regla,Mensaje).

   
*/
% quitar los asteriscos y borrar los predicados de abajo despues

 comprobar_tipo(Regla,InfrType,DefType).
 comprobar_tipos(X).
 mensaje_error(_,Num_Error,Mensaje).
 error_regla(X).
 typerror(Regla,Num_Error).
 type_Warning(Regla,Num_Error).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%    TABLA DE SIMBOLOS PARA PRUEBAS        

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% crear_tabla :-

% insert(list,class,typeCons),
% insert(list,shape,list(_)),
% insert([],class,constructor),
% insert([],arity,0),
% insert('.',class,constructor),
% insert('.',arity,2),
% insert(:,class,constructor),
% insert(:,arity,2),
% insert(tuple,class,typeCons),
% insert(',',class,constructor),
% insert(',',arity,2),
% insert(tuple,shape,tuple(_)),
% insert('Bool',class,typeCons),
% insert('Bool',shape,'Bool'),
% insert('True',class,constructor),
% insert('True',arity,0),
% insert('False',class,constructor),
% insert('False',arity,0),
% insert('Int',class,typeCons),
% insert(num,shape,num),
% insert(num,arity,1),
% insert('Constraint',class,typeCons),
% insert(constraint,shape,constraint(_)),
% insert(constraint,arity,1),
% insert(+,class,builtin),
% insert(+,type,('Int'->'Int'->'Int')),
% insert(-,class,builtin),
% insert(-,type,('Int'->'Int'->'Int')),
% insert(*,class,builtin),
% insert(*,type,('Int'->'Int'->'Int')),
% insert(/,class,builtin),
% insert(/,type,('Float'->'Float'->'Float')),
% insert(mod,class,builtin),
% insert(mod,type,('Int'->'Int'->'Int')),
% insert(div,class,builtin),
% insert(div,type,('Int'->'Int'->'Int')),
% insert(>>,class,builtin),
% insert(>>,type,('IO'(_)->'IO'(_A)->'IO'(_A))),
% insert(>>=,class,builtin),
% insert(>>=,type,('IO'(_C)->(_C->'IO'(_B))->'IO'(_B))),
% insert(done,class,function),
% insert(done,infrType,'IO'(unit)),
% insert(done,defType,'IO'(unit)),
% insert(done,definedAs,function),
% insert(return,class,function),
% insert(return,infrType,(_D->'IO'(_D))),
% insert(return,defType,(_E->'IO'(_E))),
% insert(return,definedAs,function),
% insert(getChar,class,function),
% insert(getChar,infrType,'IO'('Char')),
% insert(getChar,defType,'IO'('Char')),
% insert(getChar,definedAs,function),
% insert(putChar,class,function),
% insert(putChar,infrType,('Char'->'IO'(unit))),
% insert(putChar,defType,('Char'->'IO'(unit))),
% insert(putChar,definedAs,function),
% insert(=,class,builtin),
% insert(=,type,(_F->_F->'Constraint')),
% insert(==,class,builtin),
% insert(==,type,(_G->_G->'Bool')),
% insert(<>,class,builtin),
% insert(<>,type,(_H->_H->'Bool')),
% insert(<,class,builtin),
% insert(<,type,('Int'->'Int'->'Bool')),
% insert(>,class,builtin),
% insert(>,type,('Int'->'Int'->'Bool')),
% insert(=<,class,builtin),
% insert(=<,type,('Int'->'Int'->'Bool')),
% insert(>=,class,builtin),
% insert(>=,type,('Int'->'Int'->'Bool')),
% insert(&&,class,builtin),
% insert(&&,type,('Bool'->'Bool'->'Bool')),
% insert('||',class,builtin),
% insert('||',type,('Bool'->'Bool'->'Bool')),
% insert('@Internal',class,estado),
% insert('@Internal',rule_count,0),
% insert('@Internal',anon_var_count,0),
% insert(dup,parity,1),
% insert(dup,class,function),
% insert('@Rule1',class,rule),
% insert(dup,definedAs,function),
% insert('Nat',class,typeCons),
% insert('Nat',shape,'Nat'),
% insert('Zero',class,constructor),
% insert('Zero',arity,0),
% insert('Succ',class,constructor),
% insert('Succ',arity,1),
% insert(geq,parity,2),
% insert(geq,class,function),
% insert('@Rule2',class,rule),
% insert('@Rule3',class,rule),
% insert('@Rule4',class,rule),
% insert(geq,definedAs,function),
% insert(nat,parity,1),
% insert(nat,class,function),
% insert('@Rule5',class,rule),
% insert(nat,definedAs,function),
% insert(dupsiete,parity,0),
% insert(dupsiete,class,function),
% insert('@Rule6',class,rule),
% insert(dupsiete,definedAs,function),
% insert('@Syntax_Error',error(1),'Line'),
% insert(app,parity,2),
% insert(app,class,function),
% insert('@Rule7',class,rule),
% insert('@Rule8',class,rule),
% insert(app,definedAs,function),
% insert('@Objective',class,objective),
% insert('@Rule1',abstree,(id(dup)@[var(x)]):=(var(x)+var(x))),
% insert('@Rule2',abstree,(id(geq)@[var('_'),id('Zero')]):=id('True')),


% insert('@Rule3',abstree,
% (id(geq)@[(id('Succ')@[var(n)]),(id('Succ')@[var(m)])]):=(id(geq)@[var(n),var(m)]))
% ,
% insert('@Rule4',abstree,
% (id(geq)@[id('Zero'),(id('Succ')@[var('_')])]):=id('False')),
% insert('@Rule5',abstree,
% (id(nat)@[var(n)]):=
% (var(n)==num(0)->id('Zero')#(id('Succ')@[(id(nat)@[(var(n)@[num(1)])])]))),
% insert('@Rule6',abstree,(id(dupsiete)@[]):=(id(dup)@[num(7)])),
% insert('@Rule7',abstree,(id(app)@[id([]),var(ys)]):=var(ys)),
% insert('@Rule8',abstree,
% (id(app)@[(id('.')@[var(x),var(xs)]),var(ys)])
% :=(id('.')@[var(x),(id(app)@[var(xs),var(ys)])])),
% insert('@Objective',abstree,
% (id(app)@[[num(1),num(2),num(3)],[num(4),num(5),num(6)]])),


% insert([],type,list(_)),
% insert('.',type,(_I->list(_I)->list(_I))),
% insert(:,type,(_J->list(_J)->list(_J))),
% insert(',',type,(_K->_K->tuple(_K))),
% insert('True',type,'Bool'),
% insert('False',type,'Bool'),
% insert('Zero',type,'Nat'),
% insert('Succ',type,('Nat'->'Nat')),
% insert('@Type_data',cycle,0),
% insert('@Type_data',current_func,'Goal'),
% insert('@Type_data',current_rule,1),
% insert(app,defType,(list(_L)->list(_L)->list(_L))),
% insert(dup,type,('Int'->'Int')),
% insert(geq,type,('Nat'->'Nat'->'Bool')),
% insert(nat,type,'No Tipable'),
% insert(dupsiete,type,'Int'),
% insert(app,type,(list(A)->list(A)->list(A))).





%%%%%%%%%%%
% Predicados con utilidades
%%%%%%%%%%%

verTabla(Tabla):- findall((X,Y,Z),lookup(X,Y,Z),Tabla).

verAbstree(Lista):- findall((X,Y),lookup(X,abstree,Y),Lista).

verErrores(Lista):-
       \+ lookup('@Type_Error',_,_),Lista=['No hubo ningun error'], ! 
              ;    findall((X,Y),lookup('@Type_Error',X,Y),Lista).

verWarnings(Lista):- findall((X,Y),lookup('@Warning',X,Y),Lista).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%                 PRUEBAS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




%% pruebas de comprobar_tipo
pru0:-

% prueba correcta

insert(app,defType,(list(A)->list(A)->list(A))),

% prueba correcta con warning

insert(dus,defType,('Int'->'Int')),

% prueba que da error

insert(mal,defType,(A->'Int')).


pru1:- comprobar_tipo(app,(list(A)->list(A)->list(A)),
        (list(A)->list(A)->list(A))).

pru2 :- comprobar_tipo(dus,(A->A),('Int'->'Int')).

pru3:- comprobar_tipo(mal,(A->'Bool'),(A->'Int')).


 distinto(A,B):- \+(A=B). 

/*

%% prueba

regla(f2,[var(_B),var(_A)],
(('HP'@((app2@[num(1)])@var(_B)))@((app2@['True'])@var(_A)))).
 

tipo(app2,(list(A)->list(A)->list(A))).

*/


%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

