:- module(unref, [resolve_refs/0, free_vars/0]).

:- use_module(symtab, [get_rules/2, update_rules/2]).
:- use_module(scope,  [lookup_symbol/2]).

:- include(ast_ops).

resolve_refs :-
	get_rules(Name, Rules),
	solve_refs_list(Rules, NewRules),
	update_rules(Name, NewRules),
	fail.

resolve_refs.

solve_refs_list([], []).
solve_refs_list([R|RS], [S|SS]) :-
	solve_refs(R, S),
	!,
	solve_refs_list(RS, SS).

solve_refs(unref(X), E) :-
	(
	    lookup_symbol(X, E)
	;
	    throw(symbol_undefined(X))
	).
solve_refs(param(X), param(X)).

solve_refs(CExp, CExp) :-
	functor(CExp, _F, 0).

solve_refs(CExp, NExp) :-
	CExp =.. [F|Args],
	solve_refs(F, NF),
	solve_refs_list(Args, NArgs),
	NExp =.. [NF|NArgs].
solve_refs(E, E).

%% Rename param(x) to param(X) in all the expressions.
free_vars :- 
	get_rules(Name, Rules),
	free_rule(Rules, NewRules, _Dicc),
	update_rules(Name, NewRules),
	fail.
free_vars.

free_rule(param(X), param(V), Dicc) :-
	\+ var(X),
	!,
	member(X-V, Dicc).

free_rule(CExp, CExp, _Dicc) :-
	functor(CExp, _F, 0).

free_rule(F, NF, Dicc) :-
	F =.. [Fun|Args],
	free_rule(Fun, NFun, Dicc),
	!,
	free_rule_list(Args, NArgs, Dicc),
	NF =.. [NFun|NArgs].

free_rule(R, R, _Dicc) :- !.

free_rule_list([], [], _D).
free_rule_list([A|AA], [B|BB], Dicc) :-
	free_rule(A, B, Dicc),
	!,
	free_rule_list(AA, BB, Dicc).
	
%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

